"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createPlanton = void 0;
const delay_1 = __importDefault(require("delay"));
// @ts-expect-error -- types not available
const promise_deferred_1 = __importDefault(require("promise-deferred"));
const serialize_error_1 = require("serialize-error");
const Logger_1 = __importDefault(require("../Logger"));
const errors_1 = require("../errors");
const createEmitter_1 = require("./createEmitter");
const log = Logger_1.default.child({
    namespace: 'createPlanton',
});
const defaultCalculateDelay = () => {
    return 1000;
};
const defaultCalculateLimit = (concurrency, activeTaskInstructions) => {
    return concurrency - activeTaskInstructions.length;
};
const createPlanton = (configuration) => {
    var _a, _b;
    const { getActiveTaskInstructions, } = configuration;
    const events = (0, createEmitter_1.createEmitter)();
    const tasks = [];
    for (const inputTask of configuration.tasks) {
        log.debug('registered %s task', inputTask.name);
        for (const existingTask of tasks) {
            if (existingTask.name === inputTask.name) {
                throw new errors_1.DuplicateTaskNameError(existingTask.name);
            }
        }
        const calculateDelay = (_a = inputTask.calculateDelay) !== null && _a !== void 0 ? _a : defaultCalculateDelay;
        const calculateLimit = (_b = inputTask.calculateLimit) !== null && _b !== void 0 ? _b : defaultCalculateLimit;
        const concurrency = inputTask.concurrency === undefined ? 1 : inputTask.concurrency;
        if (concurrency < 1) {
            throw new errors_1.InvalidTaskConfigurationNameError(inputTask.name, 'Task concurrency must be greater than 0.');
        }
        const task = {
            attemptNumber: 0,
            concurrency,
            name: inputTask.name,
            schedule: inputTask.schedule,
        };
        const taskName = task.name;
        if (!taskName) {
            throw new errors_1.UnexpectedStateError('Task name cannot be empty.');
        }
        const terminate = (() => {
            let delayPromise;
            const deferredTermination = new promise_deferred_1.default();
            let active = true;
            (async () => {
                var _a;
                // eslint-disable-next-line no-unmodified-loop-condition
                while (active) {
                    const calculatedDelay = await calculateDelay((_a = task.attemptNumber) !== null && _a !== void 0 ? _a : 0);
                    if (calculatedDelay) {
                        delayPromise = (0, delay_1.default)(calculatedDelay);
                        await delayPromise;
                    }
                    if (!active) {
                        break;
                    }
                    const activeTaskInstructions = await getActiveTaskInstructions(taskName);
                    if (activeTaskInstructions.length >= concurrency) {
                        continue;
                    }
                    let taskInstructions;
                    const limit = await calculateLimit(concurrency, activeTaskInstructions);
                    if (limit < 0) {
                        const error = new errors_1.UnexpectedStateError('Limit must be greater than 0.');
                        log.error({
                            error: (0, serialize_error_1.serializeError)(error),
                            limit,
                            taskName,
                        }, 'calculateLimit an unexpected result; limit must be greater than 0');
                        events.emit('error', {
                            error,
                            taskName,
                        });
                        continue;
                    }
                    if (!Number.isInteger(limit)) {
                        const error = new errors_1.UnexpectedStateError('Limit must be an integer.');
                        log.error({
                            error: (0, serialize_error_1.serializeError)(error),
                            limit,
                            taskName,
                        }, 'calculateLimit an unexpected result; limit must be an integer');
                        events.emit('error', {
                            error,
                            taskName,
                        });
                        continue;
                    }
                    try {
                        taskInstructions = await inputTask.schedule({
                            activeTaskInstructions,
                            concurrency,
                            limit,
                            taskName,
                        });
                    }
                    catch (error) {
                        log.error({
                            error: (0, serialize_error_1.serializeError)(error),
                            taskName,
                        }, 'scheduler produced an error');
                        events.emit('error', {
                            error,
                            taskName,
                        });
                        taskInstructions = [];
                    }
                    if (!Array.isArray(taskInstructions)) {
                        events.emit('error', {
                            error: new errors_1.UnexpectedTaskInstructionsError(taskName, taskInstructions),
                            taskName,
                        });
                        log.error({
                            taskInstructions,
                            taskName,
                        }, 'scheduler produced an unexpected result; result is not array');
                        taskInstructions = [];
                    }
                    if (taskInstructions.length > limit) {
                        events.emit('error', {
                            error: new errors_1.UnexpectedTaskInstructionsError(taskName, taskInstructions),
                            taskName,
                        });
                        log.error({
                            taskInstructions,
                            taskName,
                        }, 'scheduler produced an unexpected result; instruction number is greater than the limit');
                        taskInstructions = [];
                    }
                    for (const taskInstruction of taskInstructions) {
                        if (typeof taskInstruction !== 'string') {
                            events.emit('error', {
                                error: new errors_1.UnexpectedTaskInstructionsError(taskName, taskInstructions),
                                taskName,
                            });
                            log.error({
                                taskInstructions,
                                taskName,
                            }, 'scheduler produced an unexpected result; array members are not string');
                            taskInstructions = [];
                            break;
                        }
                    }
                    if (taskInstructions.length > 0) {
                        // eslint-disable-next-line require-atomic-updates
                        task.attemptNumber = 0;
                        for (const taskInstruction of taskInstructions) {
                            events.emit('task', {
                                instruction: taskInstruction,
                                taskName,
                            });
                        }
                    }
                    else if (task.attemptNumber !== undefined) {
                        task.attemptNumber++;
                    }
                    if (!active) {
                        break;
                    }
                }
                deferredTermination.resolve();
            })();
            return () => {
                active = false;
                if (delayPromise !== undefined) {
                    // @ts-expect-error -- deferred-promise types are not available
                    delayPromise.clear();
                }
                return deferredTermination.promise;
            };
        })();
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-expect-error
        task.terminate = terminate;
        tasks.push(task);
    }
    return {
        events,
        terminate: async () => {
            await Promise.all(tasks.map(async (task) => {
                await task.terminate();
            }));
        },
    };
};
exports.createPlanton = createPlanton;
